/* ramld.c: H8 ramloader  */
/*
** Copyright (C) 2002 Martin Glunz <mg-home@gmx.de>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/


#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "h83048f.h"
#include "inh83.h"
#include "pdhw.h"

void set_vector(void *vec, void *func)
{
  *(long *)vec = (long)func | 0x5a000000; 
}

extern void vecstrt, vecend;
extern void pgmstrt;

void set_pgmvectors(void)
{
  void *dst = &vecstrt;
  void *src = &pgmstrt;
  while (dst<&vecend) {
    *(long *)dst = *(long *)src | 0x5a000000;
    dst+=4; src+=4;
  };
}


#define RXBUFLEN 128
unsigned char rxbuf[RXBUFLEN];
unsigned int rxip, rxop;

void int_rxi1(void);
void my_int_rxi1(void) __attribute((interrupt_handler));

void my_int_rxi1(void)
{
  and_ccr(0xbf);
  if (SCI1_SSR & RDRF) {
    rxbuf[rxip]=SCI1_RDR;
    if (++rxip >= RXBUFLEN) rxip=0;
  };
  SCI1_SSR &= ~(RDRF | PER | FER | ORER);
}

unsigned char *txp;
volatile short int txlen;

void txchar(unsigned char c)
{
  while (txlen>0) sleep();
  while ((SCI1_SSR & TDRE) == 0);
  SCI1_TDR=c;
  SCI1_SSR&=~TDRE;
}


void int_txi1(void);

void my_int_txi1(void) __attribute__ ((interrupt_handler));

void my_int_txi1(void)
{
  and_ccr(0xbf);
  SCI1_TDR=*txp++;
  SCI1_SSR&=~TDRE;
  if (--txlen==0) SCI1_SCR &= ~TIE;
}

void send(void *c, short int len)
{
  if (len<0) len=strlen(c);
  while (txlen>0) sleep();
  while ((SCI1_SSR & TDRE) == 0);
  txp=c; txlen=len;
  SCI1_TDR=*txp++;
  SCI1_SSR&=~TDRE;
  if (--txlen>0) SCI1_SCR |= TIE;
}

void waits(void)
{
  while (txlen>0) sleep();
}

unsigned char rxchar(void)
{
  unsigned char c;
  while (rxip==rxop) sleep();
  c=rxbuf[rxop];
  if (++rxop>=RXBUFLEN) rxop=0;
  return c;
}

unsigned long gethex(unsigned short n)
{
  unsigned long res=0;
  unsigned char c;
  while (n-->0) {
    c=rxchar();
    if (c>='a') c-=('a'-10); else
    if (c>='A') c-=('A'-10); else
    if (c>='0') c-='0';
    res = (res << 4) | (c & 0x0f);
  };
  return res;
}

/* 
set SCI_BRR here according to crystal frequency
SCI1_BRR = 3 results in 115200 baud @ 14.7456MHz xtal
SCI1_BRR = 12 for 38400 baud @ 16MHz xtal
*/

void init_ser1(void) {
  unsigned char dummy;
  SCI1_SMR =  0;  // erst Mode
  SCI1_BRR = 3;     // 115200 baud  (14.74MHz)
  SCI1_SCR = 0x30; 
  dummy = SCI1_RDR;
  SCI1_SSR &= ~(RDRF | PER | FER | ORER);
  rxip=0; rxop=0; txlen=0;
  set_vector(int_rxi1, my_int_rxi1);
  set_vector(int_txi1, my_int_txi1);
  SCI1_SCR|=RIE;
}

/*
initialize external Bus
The example Hardware has 4 external RAM chips in 8 Bit Mode (CS1 ... CS4)
and a RTC72423 @ CS5
*/

void init_bus(void) {
  // Init Bus  &  WaitStates
  ABWCR = 0xff;  // Bus Width Control Register
  ASTCR = 0x20;  // Acces Control Register (3 States fuer Uhr)
  WCR   = 0xf3;  // Wait Control Register // 3 WaitStates inserted     
  WCER  = 0xff;  // Wait-Mode in Area 5;
  BRCR  = 0xfe;
  CSCR  = 0x1f;  // Enable CS4
  P8DR  = (VPPOFF | CS1 | CS2 | CS3);
  P8DDR	= (VPPOFF | CS1 | CS2 | CS3);
  P5DDR = (A16 | A17 | A18);
  PADR  = (CS4);
  PADDR = (CS4);
  P2DDR = 0xff;
  P1DDR = 0xff;
}
        
/* 
initialize other Ports
*/
	
void init_ports(void) {
  PADR = (SRSTB | SRCLK | CS4 | CS5);
  PADDR = (SRSTB | SRCLK | CS4 | CS5);
  PBDDR = 0x00;
  P9DDR = (TXD0 | TXD1);
  //P8DR = VPPOFF;
  //P8DDR = (VPPOFF | CS1 | CS2 | CS3);
  P6DR = 0;
  P6DDR = P6DIR;
  P5DDR = (A16 | A17 | A18);
  P4DDR = 0x00;
  P2DDR = 0xff;
  P1DDR = 0xff;
}

char glb_s[80];

void main(void) __attribute__ ((noreturn));
	
void main(void) {
  unsigned short len;
  unsigned char *addr, *dump;
  unsigned char c;
  void (*strt)(void);
  short i;
  init_bus();
  //init_ports();
  init_ser1();
  //PADR |= LED0;
  and_ccr(0x3f);	// Interrupts freigeben
  txchar('X');
  while (1) {
    //send("wait for S record\n",-1);
    if (rxchar()=='S') {
      addr=NULL;
      strt=NULL;
      dump=NULL;
      c=rxchar();
      //sprintf(glb_s,"rcvd S%c\n",c);
      //send(glb_s,-1);
      switch (c) {
	case '1': len = gethex(2)-3; addr = (void *)gethex(4); break;
	case '2': len = gethex(2)-4; addr = (void *)gethex(6); break;
	case '3': len = gethex(2)-5; addr = (void *)gethex(8); break;
	case '7': len = gethex(2)-5; strt = (void *)gethex(8); break;
	case '8': len = gethex(2)-4; strt = (void *)gethex(6); break;
	case '9': len = gethex(2)-3; strt = (void *)gethex(4); break;
	case 'A': len = gethex(2); dump = (void *)gethex(4); break;
	case 'B': len = gethex(2); dump = (void *)gethex(6); break;
	case 'C': len = gethex(2); dump = (void *)gethex(8); break;
	default : len=0; break;
      };
      if (addr && len) {
        //sprintf(glb_s,"S%c addr: %08lX, len: %d\n",c,(long)addr,len);
        //send(glb_s,-1);
        while (len--) *addr++=gethex(2);
	//send("done\n",5);
      };
      if (dump) {
        sprintf(glb_s,"%08lX: ",(long)dump); send(glb_s,-1); waits();
	for (i=0; i<16; i++) {
	  sprintf(glb_s,"%02X ",*dump++); send(glb_s,-1); waits();
	};
	send("\n",1);
      };
      if (strt) {
        gethex(2); rxchar();
        //sprintf(glb_s,"S%c start: %08lX\n",c,(long)strt);
        //send(glb_s,-1); waits();
  	or_ccr(0xc0);	// disable interrupts
	set_pgmvectors();
	strt();
      };
    };
  };
}
